from builtins import bytes
import struct
from . import nsis2, nsis3
from .. import fileform

SYSVAR_NAMES = {
    20: 'COMMANDLINE',
    21: 'INSTALLDIR',
    22: 'OUTPUTDIR',
    23: 'EXEDIR',
    24: 'LANGUAGE',
    25: 'TEMPDIR',
    26: 'PLUGINSDIR',
    27: 'EXEPATH',
    28: 'EXEFILE',
    29: 'HWNDPARENT',
    30: 'CLICKNEXT',
}

ESCAPE_MAP = {
    0x9: '$\\t',
    0xa: '$\\n',
    0xd: '$\\r',
    0x22: '$\\"',
    0x24:'$$',
}

class Symbol(object):
    def is_reg(self):
        return False
    def is_var(self):
        return False
    def is_nvar(self):
        return False
    def is_lang_code(self):
        return False
    def is_shell(self):
        return False
    def is_string(self):
        return False

class NVar(Symbol):
    def __init__(self, nvar):
        self.nvar = nvar

    def __str__(self):
        if self.nvar in SYSVAR_NAMES:
            return '$' + SYSVAR_NAMES[self.nvar]
        elif self.nvar < 10:
            return '$' + str(self.nvar)
        elif self.nvar < 20:
            return '$R' + str(self.nvar - 10)
        else:
            return '$__var{}__'.format(self.nvar)

    def is_nvar(self):
        return True

    def is_reg(self):
        return self.nvar < 20

    def is_var(self):
        return self.nvar >= 20

class LangCode(Symbol):
    def __init__(self, nlang):
        self.nlang = nlang

    def __str__(self):
        return '$(LangString{})'.format(self.nlang)

    def is_lang_code(self):
        return True

class Shell(Symbol):
    def __init__(self, param1, param2):
        self.param1 = param1
        self.param2 = param2

    def __str__(self):
        return '$__SHELL_{}_{}__'.format(self.param1, self.param2)

    def is_shell(self):
        return True

class String(Symbol, str):
    def is_string(self):
        return True

def _symbolize(block, offset, code_helper):
    """ Decode special characters found in NSIS strings. """

    symbols = []
    cur_string = ""
    data = bytes(block[offset:offset + fileform.NSIS_MAX_STRLEN])
    i = 0
    while i < len(data):
        c = data[i]
        i += 1

        if c == 0:
            break

        if code_helper.is_code(c):
            if cur_string:
                symbols.append(String(cur_string))
                cur_string = ""

            param1 = data[i]
            param2 = data[i+1]
            param = (param1 & 0x7f) | ((param2 & 0x7f) << 7)

            i += 2
            if c == code_helper.NS_SHELL_CODE:
                symbols.append(Shell(param1, param2))
            elif c == code_helper.NS_VAR_CODE:
                symbols.append(NVar(param))
            elif c == code_helper.NS_LANG_CODE:
                symbols.append(LangCode(param))
        elif c == code_helper.NS_SKIP_CODE:
            cur_string += data[i]
            i += 1
        elif c in ESCAPE_MAP:
            cur_string += ESCAPE_MAP[c]
        else:
            cur_string += chr(c)

    if cur_string:
        symbols.append(String(cur_string))

    return symbols, i

def symbolize(block, offset, version='3'):
    if version == '3':
        return _symbolize(block, offset, nsis3)
    elif version == '2':
        return _symbolize(block, offset, nsis2)
    else:
        raise Exception('Unknown NSIS version: ' + repr(version))

def decode(block, offset=0, version='3'):
    symbols, i = symbolize(block, offset, version)
    string = ''
    for s in symbols:
        string += str(s)
    return string, i




import nrs
import idaapi
import string
import nrs
from nrs import fileform, nsisfile

PTR_NONE = 0xffffffff

BLOCKS = [
    ('PAGES', fileform.NB_PAGES, 'DATA'),
    ('SECTIONS', fileform.NB_SECTIONS, 'DATA'),
    ('ENTRIES', fileform.NB_ENTRIES, 'CODE'),
    ('STRINGS', fileform.NB_STRINGS, 'DATA'),
    ('LANGTABLES', fileform.NB_LANGTABLES, 'DATA'),
    ('CTLCOLORS', fileform.NB_CTLCOLORS, 'DATA'),
    ('BGFONT', fileform.NB_BGFONT, 'DATA'),
    ('DATA', fileform.NB_DATA, 'DATA'),
]

allowed_name_char = string.ascii_letters + string.digits + '$'
def canonize_name(name):
    """ Limit names to a subset of ascii character. """
    return str(''.join([c if c in allowed_name_char else '_' for c in name]))

def align(addr):
    return (addr + 0xfff) & 0xfffff000

def accept_file(li, n):
    li.seek(0)
    if n == 0 and fileform._find_firstheader(li):
        return "NSIS (NullSoft Installer)"
    return 0

def load_file(li, netflags, format):
    nsis = nsisfile.NSIS.from_path(idaapi.get_input_file_path())

    # Create NSIS netnode.
    nsis_netnode = idaapi.netnode('$ NSIS', 0, True)
    nsis_netnode.hashset('VERSION_MAJOR', nsis.version_major)
    nsis_netnode.hashset('VERSION_MINOR', nsis.version_minor)

    # Create blocks segments.
    for name, n, sclass in BLOCKS:
        offset = nsis.header.blocks[n].offset
        if offset == 0:
            continue
        content = nsis.block(n)
        # Create block segment
        seg = idaapi.segment_t()
        seg.startEA = offset
        seg.endEA = offset + len(content)
        idaapi.add_segm_ex(seg, name, sclass, 0)
        idaapi.mem2base(content, offset)

    # Add one virtual segment to hold variables.
    var_seg = idaapi.segment_t()
    var_start = align(nsis.size())
    var_seg.startEA = var_start
    var_seg.endEA = var_start + 0x1000 # Size chosen arbitrarily, should be enough.
    idaapi.add_segm_ex(var_seg, 'VARS', 'BSS', 0)
    # Create standard vars.
    for i, v in enumerate(nrs.strings.SYSVAR_NAMES.values()):
        idaapi.do_name_anyway(var_seg.startEA + i + 20, '$' + v)

    code_base = nsis.header.blocks[fileform.NB_ENTRIES].offset
    # Create sections functions.
    for i, section in enumerate(nsis.sections):
        if section.code == PTR_NONE:
            continue
        name = nsis.get_string(section.name_ptr)
        if not name:
            name = '_section' + str(i)
        ea = code_base + nrs.entry_to_offset(section.code)
        cname = canonize_name(name)
        AddEntryPoint(ea, ea, cname, 1)

    # Mark pages handlers.
    for i, page in enumerate(nsis.pages):
        for fn in ['prefunc', 'showfunc', 'leavefunc']:
            addr = getattr(page, fn)
            if addr != PTR_NONE:
                name = '_page_{}_{}'.format(i, fn)
                ea = code_base + nrs.entry_to_offset(addr)
                AddEntryPoint(ea, ea, name, 1)

    # Mark installer handlers.
    for event in ['Init', 'InstSuccess', 'InstFailed', 'UserAbort', 'GUIInit',
                  'GUIEnd', 'MouseOverSection', 'VerifyInstDir', 'SelChange',
                  'RebootFailed']:
        addr = getattr(nsis.header, 'code_on'+event)
        if addr != PTR_NONE:
            name = '_on' + event
            ea = code_base + nrs.entry_to_offset(addr)
            AddEntryPoint(ea, ea, name, 1)

    # Create strings.
    """
    strings_data = nsis.block(fileform.NB_STRINGS)
    strings_off = nsis.header.blocks[fileform.NB_STRINGS].offset
    i = 0
    while i < len(strings_data):
        decoded_string, length = \
            nrs.strings.decode(strings_data, i, nsis.version_major)
        decoded_string = str(decoded_string)
        string_name = canonize_name(decoded_string)
        idaapi.make_ascii_string(strings_off + i, length, ASCSTR_C)
        idaapi.set_cmt(strings_off + i, decoded_string, True)
        idaapi.do_name_anyway(strings_off + i, string_name)
        i += length
    #"""


    # Set processor to nsis script.
    SetProcessorType("nsis", SETPROC_ALL|SETPROC_FATAL)
    return 1




from idaapi import *
import ntpath
import idaapi
import struct
import string
import nrs

allowed_name_char = string.ascii_letters + string.digits + '$'
def canonize_name(name):
    """ Limit names to a subset of ascii character. """
    return str(''.join([c if c in allowed_name_char else '_' for c in name]))

def str_to_number(sym):
    if not sym.is_string():
        return None

    try:
        if sym[:2] == '0x':
            return int(sym, 16)
        elif sym[0] == '0':
            return int(sym, 8)
        else:
            return int(sym)
    except:
        return None

STR_LANG_FLAG = 1 << 31
OP_SIZE = 4
INST_SIZE = 7 * OP_SIZE

class NsisProcessor(processor_t):
    """ NSIS Processor class used by IDA to disassemble and analyze NSIS code. """
    # IDP id (Chosen arbitrarily > 0x8000).
    id = 0x8513

    # Processor features.
    flag = PR_USE32 | PR_DEFSEG32 | PR_RNAMESOK | PRN_HEX | PR_NO_SEGMOVE

    # Bits in byte for code segments.
    cnbits = 8

    #  Bits in byte for non-code segments.
    dnbits = 8

    # Short processor names.
    psnames = ['nsis']

    # Long processor names.
    plnames = ['NSIS Script Byte code']

    # Size of a segment registrer in bytes.
    segreg_size = 0

    # Array or 'return' instruction opcodes.
    retcodes = [struct.pack('<I',1) + struct.pack('<I', 0) * 6]

    # First's instruction icode.
    instruc_start = 0

    # Size of long double.
    tbyte_size = 0

    # Assembler features.
    assembler = {
        # Assembler flags.
        'flag': ASB_BINF3 | ASH_HEXF3 | ASO_OCTF1,

        # User defined flags.
        'uflag': 0,

        # Assembler name.
        'name': 'NSIS Script Byte code assembler',

        # org directive.
        'origin': 'org',

        # end directive.
        'end': 'end',

        # Comment string.
        'cmnt': ';',

        # ASCII string delimiter.
        'ascsep': '"',

        # ASCII char constant delimiter.
        'accsep': "'",

        # ASCII special chars.
        'esccodes': '"\'',

        # Data representation.
        'a_ascii': 'db',
        'a_byte': 'db',
        'a_word': 'dw',
        'a_dword': 'dd',

        'a_dups': '#d dup(#v)',
        'a_bss': '%s dup ?',
        'a_seg': 'seg',
        'a_curip': '$',
        'a_public': 'public',
        'a_weak': 'weak',
        'a_extrn': 'extrn',
        'a_comdef': '',
        'a_align': 'align',
        'lbrace': '(',
        'rbrace': ')',
        'a_mod': '%',
        'a_band': '&',
        'a_bor': '|',
        'a_xor': '^',
        'a_bnot': '~',
        'a_shl': '<<',
        'a_shr': '>>',
        'a_sizeof_fmt': 'size %s',
    } # Assembler.

    INTOP_SYM = ['+','-','*','/','|','&','^','!','||','&&','%','<<','>>']

    # NSIS specific flags.
    FLo_IntOp =       0x01
    FLo_PluginCall =  0x02

    FLa_CheckNoFlow = 0x04
    FLa_NoFlow =      0x08
    FLa_StackArgs =   0x10

    def rebase_string_addr(self, addr):
        if addr & STR_LANG_FLAG:
            return addr
        seg = get_segm_by_name('STRINGS')
        return addr + seg.startEA

    def rebase_var_addr(self, addr):
        seg = get_segm_by_name('VARS')
        return addr + seg.startEA

    def rebase_code_entry(self, entry):
        seg = get_segm_by_name('ENTRIES')
        return nrs.entry_to_offset(entry) + seg.startEA

    def get_string_symbols(self, addr):
        seg = get_segm_by_name('STRINGS')
        if not seg.contains(addr):
            return None

        seg = get_segm_by_name('STRINGS')
        maxlen = min(seg.endEA - addr, nrs.fileform.NSIS_MAX_STRLEN)
        data = GetManyBytes(addr, maxlen)
        symbols, _ = nrs.strings.symbolize(data, 0, self.nsis_version)
        return symbols

    def get_string(self, addr):
        seg = get_segm_by_name('STRINGS')
        if not seg.contains(addr):
            return None, 0

        maxlen = min(seg.endEA - addr, nrs.fileform.NSIS_MAX_STRLEN)
        data = GetManyBytes(addr, maxlen)
        string, l = nrs.strings.decode(data, 0, self.nsis_version)
        return str(string), l

    def read_params(self):
        return [ua_next_long() for _ in range(6)]

    def decode_plugin_call(self, opcode, params):
        if opcode != self.itype_CALL: return
        if ua_next_long() != self.itype_EXTRACTFILE: return
        self.read_params()
        if ua_next_long() != self.itype_SETFLAG: return
        self.read_params()
        argn = 0
        while True:
            opcode = ua_next_long()
            if opcode == self.itype_PUSHPOP and \
                    self.read_params()[1:3] == [0,0]:
                argn += 1
            elif opcode == self.itype_REGISTERDLL:
                # Plugin call!
                params = self.read_params()
                self.cmd.itype = self.itype_PLUGINCALL
                self.op_str(self.cmd.Op1, params[0])
                self.op_str(self.cmd.Op2, params[1])
                self.op_imm(self.cmd.Op3, argn)
                self.cmd.Op3.specval |= self.FLa_StackArgs
                self.cmd.auxpref |= self.FLo_PluginCall
                return True
            else:
                return

    def get_plugin_call_args(self, cmd, op):
        first = cmd.ea + INST_SIZE * 3
        return [self.rebase_string_addr(Dword(first + OP_SIZE + i*INST_SIZE))
                    for i in range(op.value)]

    def get_frame_retsize(self):
        return 4

    def header(self):
        return 'NSIS Script V' + str(self.nsis_version)

    def ana(self):
        """ Decode NSIS instruction. """
        opcode = ua_next_long()
        params = self.read_params()

        if self.decode_plugin_call(opcode, params):
            return self.cmd.size
        self.cmd.size = INST_SIZE

        if opcode < len(self.itable):
            ins = self.itable[opcode]
        else:
            ins = self.itable[0]

        # Decode "virtual instruction" (eg. PUSHPOP -> PUSH/POP/EXCH)
        if ins.v:
            opcode = ins.v(opcode, params)
            ins = self.itable[opcode]

        self.cmd.itype = opcode
        self.cmd.auxpref |= ins.ap
        return self.cmd.size if self.decode(ins.d, params) else 0

    def handle_string(self, offb, op, addr):
        sym_addr = addr
        symbols = self.get_string_symbols(addr)
        if symbols:
            for i, symbol in enumerate(symbols):
                if symbol.is_var():
                    var_addr = self.rebase_var_addr(symbol.nvar)
                    ua_add_dref(offb, var_addr, dr_R)
                    sym_addr += 4
                elif symbol.is_string():
                    n = str_to_number(symbol)
                    if n is None:
                        ua_add_dref(offb, sym_addr, dr_R)
                        string_name = canonize_name(symbol)
                        idaapi.make_ascii_string(sym_addr, len(symbol), ASCSTR_C)
                        idaapi.do_name_anyway(sym_addr, string_name[:15])
                    sym_addr += len(symbol)
                else:
                    sym_addr += 4

    def handle_operand(self, op, isRead):
        dref_flag = dr_R if isRead else dr_W
        offb = (op.n+1)*OP_SIZE

        if op.type == o_mem:
            if op.dtyp == dt_string:
                self.handle_string(offb, op, op.addr)
            else:
                ua_add_dref(offb, op.addr, dref_flag)
        elif op.type == o_near:
            if self.cmd.itype == self.itype_CALL:
                fl = fl_CN
            else:
                fl = fl_JN
            ua_add_cref(offb, op.addr, fl)
        elif op.type == o_imm and op.specval & self.FLa_StackArgs:
            for arg in self.get_plugin_call_args(self.cmd, op):
                self.handle_string(0, op, arg)

    def emu(self):
        """ Emulate instruction behavior. """
        feature = self.cmd.get_canon_feature()

        if feature & CF_USE1:
            self.handle_operand(self.cmd.Op1, 1)
        if feature & CF_USE2:
            self.handle_operand(self.cmd.Op2, 1)
        if feature & CF_USE3:
            self.handle_operand(self.cmd.Op3, 1)
        if feature & CF_USE4:
            self.handle_operand(self.cmd.Op4, 1)
        if feature & CF_USE5:
            self.handle_operand(self.cmd.Op5, 1)
        if feature & CF_USE6:
            self.handle_operand(self.cmd.Op6, 1)

        if feature & CF_CHG1:
            self.handle_operand(self.cmd.Op1, 0)
        if feature & CF_CHG2:
            self.handle_operand(self.cmd.Op2, 0)
        if feature & CF_CHG3:
            self.handle_operand(self.cmd.Op3, 0)
        if feature & CF_CHG4:
            self.handle_operand(self.cmd.Op4, 0)
        if feature & CF_CHG5:
            self.handle_operand(self.cmd.Op5, 0)
        if feature & CF_CHG6:
            self.handle_operand(self.cmd.Op6, 0)

        if feature & CF_JUMP:
            QueueSet(Q_jumps, self.cmd.ea)

        # Add flow cref.
        noFlow = self.get_auxpref() & self.FLa_NoFlow
        if not (feature & CF_STOP or noFlow):
            ua_add_cref(0, self.cmd.ea + self.cmd.size, fl_F)

        return 1

    def out(self):
        """ Output instruction in textform. """
        buf = idaapi.init_output_buffer(1024)

        if self.cmd.auxpref & self.FLo_PluginCall:
            lib,_ = self.get_string(self.cmd[0].addr)
            fn,_ = self.get_string(self.cmd[1].addr)
            lib = ntpath.splitext(ntpath.basename(lib))[0]
            out_line('{}::{}'.format(lib, fn), COLOR_INSN)
            OutChar(' ')
            out_one_operand(2)
        else:
            OutMnem(12)
            for i, op in ((i, self.cmd[i]) for i in range(6)):
                if op.type == o_void:
                    break
                if i > 0:
                    out_symbol(',')
                    OutChar(' ')
                out_one_operand(i)

        term_output_buffer()
        cvar.gl_comm = 1
        MakeLine(buf)

    def out_str(self, op, addr):
        symbols = self.get_string_symbols(addr)
        # Translated string.
        if symbols is None:
            self.out_name_addr(op, addr)
        elif not symbols:
            out_line('""', COLOR_STRING)
        else:
            for i, symbol in enumerate(symbols):
                if symbol.is_reg():
                    out_register(self.regNames[symbol.nvar])
                elif symbol.is_var():
                    var_addr = self.rebase_var_addr(symbol.nvar)
                    out_name_expr(op, var_addr, var_addr)
                else:
                    n = str_to_number(symbol)
                    if n is None:
                        out_line('"' + str(symbol) + '"', COLOR_STRING)
                    else:
                        out_long(n, 16)

    def out_name_addr(self, op, addr):
        r = out_name_expr(op, addr, BADADDR)
        if not r:
            out_tagon(COLOR_ERROR)
            OutLong(op.addr, 16)
            out_tagoff(COLOR_ERROR)
            QueueSet(Q_noName, self.cmd.ea)

    def outop(self, op):
        """ Output instruction's operand in textform. """

        if op.type == o_reg:
            out_register(self.regNames[op.reg])
        elif op.type == o_imm:
            if op.specval & self.FLo_IntOp:
                out_line(self.INTOP_SYM[op.value], COLOR_SYMBOL)
            elif op.specval & self.FLa_StackArgs:
                args = self.get_plugin_call_args(self.cmd, op)
                for i, arg in enumerate(args):
                    if i > 0:
                        out_symbol(',')
                        OutChar(' ')
                    self.out_str(op, arg)
            else:
                OutValue(op, OOFW_IMM | OOF_SIGNED)
        elif op.type == o_near:
            self.out_name_addr(op, op.addr)
        elif op.type == o_mem:
            if op.dtyp == dt_string:
                self.out_str(op, op.addr)
            else:
                self.out_name_addr(op, op.addr)
        else:
            return False
        return True

    # Instruction decoding functions.
    def op_str(self, op, addr):
        # Some string are used as numbers, register or nvar.
        addr = self.rebase_string_addr(addr)
        symbols = self.get_string_symbols(addr)
        if symbols and len(symbols) == 1:
            symbol = symbols[0]
            if symbol.is_nvar():
                return self.op_var(op, symbol.nvar)
            if symbol.is_string():
                n = str_to_number(symbol)
                if n is not None:
                    return self.op_imm(op, n)

        op.type = o_mem
        op.dtyp = dt_string
        op.addr = addr


    def op_imm(self, op, imm):
        op.type = o_imm
        op.dtyp = dt_dword
        op.value = imm

    def op_var(self, op, x):
        if x < 20:
            op.type = o_reg
            op.reg = x
        elif x == 0xffffffff:
            op.type = o_imm
            op.dtyp = dt_dword
            op.value = -1
        else:
            op.type = o_mem
            op.dtyp = dt_byte
            op.addr = self.rebase_var_addr(x)

    def op_jmp(self, op, addr):
        op.type = o_near
        op.addr = self.rebase_code_entry(addr-1)

    def op_void(self, op):
        op.type = o_void

    def decode(self, fmt, params):
        if fmt == '':
            self.op_void(self.cmd.Op1)
            return True

        noFlow = False
        # Most instruction with jumps can lead to no flow reference if no
        # jumps address are set to 0.
        if 'J' in fmt and self.cmd.auxpref & self.FLa_CheckNoFlow:
            noFlow = True

        for i, (c,p) in enumerate(zip(fmt, params)):
            op = self.cmd[i]
            if c == 'I':
                self.op_imm(op, p)
            elif c == 'S':
                self.op_str(op, p)
            elif c == 'V':
                self.op_var(op, p)
            elif c == 'J':
                if p > 0:
                    self.op_jmp(op, p)
                else:
                    self.op_imm(op, p)
                    noFlow = False
            elif c == 'O': # Math operand.
                self.op_imm(op, p)
                op.specval |= self.FLo_IntOp
            elif c == '2': # SendMessage bitshift one of its operant.
                self.op_imm(op, p >> 2)
            else:
                raise Exception('Unknown format flag: ' + c)

        if noFlow:
            self.cmd.auxpref |= self.FLa_NoFlow
        return True

    def virt_pushpop(self, opcode, params):
        if params[1]:
            return self.itype_POP
        elif params[2]:
            return self.itype_EXCH
        else:
            return self.itype_PUSH

    def virt_setflag(self, opcode, params):
        if params[0] == 2 and params[1] == 0xac:
            return self.itype_CLEARERRORS
        return opcode

    def virt_ifflag(self, opcode, params):
        if params[1] == 0 and params[2] == 2 and params[3] == 0:
            return self.itype_IFERRORS
        return opcode

    def virt_strcpy(self, opcode, params):
        if params[2] == 0 and params[3] == 0:
            return self.itype_ASSIGNVAR
        return opcode

    def virt_showwindow(self, opcode, params):
        if params[2]:
            return self.itype_HIDEWINDOW
        elif params[3]:
            return self.itype_ENABLEWINDOW
        return opcode

    def virt_delreg(self, opcode, params):
        if params[4]:
            return self.itype_DELETEREGKEY
        return self.itype_DELETEREGVALUE

    def virt_regenum(self, opcode, params):
        if params[4]:
            return self.itype_REGENUMKEY
        return self.itype_REGENUMVALUE

    def virt_fwrite(self, opcode, params):
        if params[2]:
            return self.itype_FILEWRITEBYTE
        return self.itype_FILEWRITE

    def virt_fread(self, opcode, params):
        if params[3]:
            return self.itype_FILEREADBYTE
        return self.itype_FILEREAD

    def virt_log(self, opcode, params):
        if params[0]:
            return self.itype_LOGSET
        return self.itype_LOGTEXT

    def init_instructions(self):
        class idef:
            def __init__(self, name, cf=0, d='', v=None, ap=0):
                self.name = name
                self.cf = cf
                self.d = d
                self.v = v
                self.ap = ap

        i_invalid = idef(name='INVALID')
        def notimplemented(n):
            return idef(name='NotImplemented_' + hex(n))

        self.itable = [
            i_invalid, # 0x00
            idef(name='Return', d='', cf=CF_STOP), # 0x01
            idef(name='Jmp', d='J', cf=CF_USE1, ap=self.FLa_CheckNoFlow), # 0x02
            idef(name='Abort', d='I', cf=CF_USE1|CF_STOP), # 0x03
            idef(name='Quit', cf=CF_STOP), #0x04
            idef(name='Call', d='J', cf=CF_USE1|CF_CALL), # 0x05
            idef(name='UpdateText', d='S', cf=CF_USE1), # 0x06
            idef(name='Sleep', d='I', cf=CF_USE1), # 0x07
            idef(name='BringToFront'), # 0x08
            idef(name='ChDetailsView', d='SS', cf=CF_USE1|CF_USE2), # 0x09
            idef(name='SetFileAttributes', d='SI', cf=CF_USE1|CF_USE2), # 0x0a
            idef(name='CreateDir', d='SI', cf=CF_USE1|CF_USE2), # 0x0b
            idef(name='IfFileExists', d='SJJ', cf=CF_USE1|CF_USE2|CF_USE3, ap=self.FLa_CheckNoFlow), # 0x0c
            idef(name='SetFlag', d='IS', v=self.virt_setflag, cf=CF_USE1|CF_USE2), # 0x0d
            idef(name='IfFlag', d='JJII',v=self.virt_ifflag, cf=CF_USE1|CF_USE2|CF_USE3|CF_USE4, ap=self.FLa_CheckNoFlow), # 0x0e
            idef(name='GetFlag', d='VI', cf=CF_CHG1|CF_USE2), # 0x0f
            idef(name='Rename', d='SSIS', cf=CF_CHG1|CF_USE2|CF_USE3|CF_USE4), # 0x10
            idef(name='GetFullPathName', d='SVI', cf=CF_USE1|CF_CHG2|CF_USE3), # 0x11
            idef(name='SearchPath', d='VS', cf=CF_CHG1|CF_USE2), # 0x12
            idef(name='GetTempFilename', d='VS', cf=CF_CHG1|CF_USE2), # 0x13
            idef(name='ExtractFile', d='ISIIII', cf=CF_USE1|CF_USE2|CF_USE3|CF_USE4|CF_USE5|CF_USE6), # 0x14
            idef(name='DeleteFile', d='SI', cf=CF_USE1|CF_USE2), # 0x15
            idef(name='MessageBox', d='ISIJIJ', cf=CF_USE1|CF_USE2|CF_USE3|CF_USE4|CF_USE5), # 0x16
            idef(name='RmDir', d='SI', cf=CF_USE1|CF_USE2), # 0x17
            idef(name='StrLe', d='VS', v=self.virt_setflag, cf=CF_CHG1|CF_USE2), # 0x18
            idef(name='StrCpy', d='VSSS', v=self.virt_strcpy, cf=CF_CHG1|CF_USE2|CF_USE3|CF_USE4), # 0x19
            idef(name='StrCmp', d='SSJJI', cf=CF_USE1|CF_USE2|CF_USE3|CF_USE4|CF_USE5, ap=self.FLa_CheckNoFlow), # 0x1a
            idef(name='ReadEnv', d='VSI', cf=CF_CHG1|CF_USE2|CF_USE3), # 0x1b
            idef(name='IntCmp', d='SSJJJI', cf=CF_USE1|CF_USE2|CF_USE3|CF_USE4|CF_USE5|CF_USE6, ap=self.FLa_CheckNoFlow), # 0x1c
            idef(name='IntOp', d='VSSO', cf=CF_CHG1|CF_USE2|CF_USE3|CF_USE4), # 0x1d
            idef(name='IntFmt', d='VSS', cf=CF_CHG1|CF_USE2|CF_USE3), # 0x1d
            idef(name='PushPop', v=self.virt_pushpop), # 0x1f
            idef(name='FindWindow', d='VSSSS', cf=CF_CHG1|CF_USE2|CF_USE3|CF_USE4|CF_USE5), # 0x20
            idef(name='SendMessage', d='VSSSS2', cf=CF_CHG1|CF_USE2|CF_USE3|CF_USE4|CF_USE5|CF_USE6), # 0x21
            idef(name='IsWindow', d='SJJ', cf=CF_USE1|CF_USE2|CF_USE3, ap=self.FLa_CheckNoFlow), # 0x22
            idef(name='GetDlgItem', d='VSS', cf=CF_CHG1|CF_USE2|CF_USE3), # 0x23
            idef(name='SetCtlColors', d='SI', cf=CF_USE1|CF_USE2), # 0x24
            idef(name='SetBrandingImage', d='SII', cf=CF_USE1|CF_USE2), # 0x25
            idef(name='CreateFont', d='VSSSI', cf=CF_CHG1|CF_USE2|CF_USE3|CF_USE4|CF_USE5), # 0x26
            idef(name='ShowWindow', d='SS', v=self.virt_showwindow, cf=CF_USE1|CF_USE2), # 0x27
            idef(name='ShellExec', d='SSSS', cf=CF_USE1|CF_USE2|CF_USE3|CF_USE4), # 0x28
            idef(name='Execute', d='SII', cf=CF_USE1|CF_USE2|CF_USE3), # 0x29
            idef(name='GetFileTime', d='VVS', cf=CF_CHG1|CF_CHG2|CF_USE3), # 0x2a
            idef(name='GetDLLVersion', d='VVS', cf=CF_CHG1|CF_CHG2|CF_USE3), # 0x2b
            idef(name='RegisterDLL', d='SSSI', cf=CF_USE1|CF_USE2|CF_USE3|CF_USE4), # 0x2c
            idef(name='CreateShortcut', d='SSSSS', cf=CF_USE1|CF_USE2|CF_USE3|CF_USE4|CF_USE5), # 0x2d
            idef(name='CopyFiles', d='SSS', cf=CF_USE1|CF_USE2|CF_USE3), # 0x2e
            idef(name='Reboot'), # 0x2f
            idef(name='WriteIni', d='SSSS', cf=CF_USE1|CF_USE2|CF_USE3|CF_USE4), # 0x30
            idef(name='ReadIni', d='VSSS', cf=CF_CHG1|CF_USE2|CF_USE3|CF_USE4), # 0x31
            idef(name='DeleteRegKey', d='ISSS', v=self.virt_delreg, cf=CF_USE1|CF_USE2|CF_USE3|CF_USE4), # 0x32
            idef(name='WriteRegValue', d='ISSII', cf=CF_USE1|CF_USE2|CF_USE3|CF_USE4|CF_USE5), # 0x33
            idef(name='ReadRegValue', d='VISSI', cf=CF_CHG1|CF_USE2|CF_USE3|CF_USE4|CF_USE5), # 0x34
            idef(name='RegEnumKey', d='VISS', v=self.virt_regenum, cf=CF_CHG1|CF_USE2|CF_USE3|CF_USE4), # 0x35
            idef(name='FileClose', d='V', cf=CF_USE1), # 0x36
            idef(name='FileOpen', d='VIIS', cf=CF_CHG1|CF_USE2|CF_USE3|CF_USE4), # 0x37
            idef(name='FileWrite', d='VS', v=self.virt_fwrite, cf=CF_USE1|CF_USE2), # 0x38
            idef(name='FileRead', d='VVS', v=self.virt_fread, cf=CF_USE1|CF_CHG2|CF_USE3), # 0x39
            idef(name='FileSeek', d='VVSI', cf=CF_USE1|CF_CHG2|CF_USE3|CF_USE4), # 0x3a
            idef(name='FindClose', d='V', cf=CF_USE1), # 0x3b
            idef(name='FindNext', d='VV', cf=CF_CHG1|CF_USE2), # 0x3c
            idef(name='FindFirst', d='VVS', cf=CF_CHG1|CF_CHG2|CF_USE3), # 0x3d
            idef(name='WriteUninstaller', d='SIIS', cf=CF_USE1|CF_USE2|CF_USE3|CF_USE4), # 0x3e
            idef(name='LogText', d='S', v=self.virt_log, cf=CF_USE1), # 0x3f
            idef(name='SectionSet', d='SII', cf=CF_USE1|CF_USE2|CF_USE3), # 0x40
            idef(name='InstTypeSet', d='SIII', cf=CF_USE1|CF_USE2|CF_USE3|CF_USE4), # 0x41
            idef(name='GetLabelAddr'), # 0x42
            idef(name='GetFunctionAddr'), # 0x43
            idef(name='LockWindow', d='I', cf=CF_USE1), # 0x44
        ]

        self.itable += [
            idef(name='Push', d='S', cf=CF_USE1),
            idef(name='Pop', d='V', cf=CF_CHG1),
            idef(name='Exch', d='I', cf=CF_USE1|CF_CHG1),
            idef(name='ClearErrors'),
            idef(name='IfErrors', d='J', cf=CF_USE1),
            idef(name='AssignVar', d='VS', cf=CF_CHG1|CF_USE2),
            idef(name='EnableWindow', d='SS', cf=CF_USE1|CF_USE2),
            idef(name='HideWindow', d='SS', cf=CF_USE1|CF_USE2),
            idef(name='DeleteRegValue', d='ISSS', cf=CF_USE1|CF_USE2|CF_USE3|CF_USE4),
            idef(name='RegEnumValue', d='VISS', cf=CF_CHG1|CF_USE2|CF_USE3|CF_USE4),
            idef(name='FileWriteByte', d='VS', cf=CF_USE1|CF_USE2),
            idef(name='FileReadByte', d='VV', cf=CF_USE1|CF_CHG2),
            idef(name='LogSet', d='I', cf=CF_USE1),
            idef(name='PluginCall', cf=CF_USE1|CF_USE2|CF_USE3),
        ]

        # Now create an instruction table compatible with IDA processor module requirements
        instructions = []
        for i, x in enumerate(self.itable):
            d = dict(name=x.name, feature=x.cf)
            instructions.append(d)
            setattr(self, 'itype_' + x.name.upper(), i)

        # icode of the last instruction + 1
        self.instruc_end = len(instructions) + 1

        # Array of instructions
        self.instruc = instructions

        # Icode of return instruction. It is ok to give any of possible return
        # instructions
        self.icode_return = self.itype_RETURN

    def init_registers(self):
        """
        This function parses the register table and creates corresponding
        ireg_XXX constants
        """
        self.regNames = sorted([x for n in range(10) for x in ('$'+str(n), '$R'+str(n))])
        self.regNames += ['CS','DS'] # Fake segment registers.

        # Create the ireg_XXXX constants
        for i, name in enumerate(self.regNames):
            setattr(self, 'ireg_' + name, i)

        # Segment register information (use virtual CS and DS registers if your
        # processor doesn't have segment registers):
        self.regFirstSreg = self.ireg_CS
        self.regLastSreg  = self.ireg_DS

        # number of CS register
        self.regCodeSreg = self.ireg_CS

        # number of DS register
        self.regDataSreg = self.ireg_DS

    def __init__(self):
        idaapi.processor_t.__init__(self)
        self.PTRSZ = 4
        self.init_instructions()
        self.init_registers()
        self.nsis_netnode = netnode('$ NSIS')
        self.nsis_version = self.nsis_netnode.hashstr('VERSION_MAJOR')

def PROCESSOR_ENTRY():
    return NsisProcessor()




# This file was automatically generated by SWIG (http://www.swig.org).
# Version 3.0.7
#
# Do not make changes to this file unless you know what you are doing--modify
# the SWIG interface file instead.





from sys import version_info
if version_info >= (2, 6, 0):
    def swig_import_helper():
        from os.path import dirname
        import imp
        fp = None
        try:
            fp, pathname, description = imp.find_module('_bzlib', [dirname(__file__)])
        except ImportError:
            import _bzlib
            return _bzlib
        if fp is not None:
            try:
                _mod = imp.load_module('_bzlib', fp, pathname, description)
            finally:
                fp.close()
            return _mod
    _bzlib = swig_import_helper()
    del swig_import_helper
else:
    import _bzlib
del version_info
try:
    _swig_property = property
except NameError:
    pass  # Python < 2.2 doesn't have 'property'.


def _swig_setattr_nondynamic(self, class_type, name, value, static=1):
    if (name == "thisown"):
        return self.this.own(value)
    if (name == "this"):
        if type(value).__name__ == 'SwigPyObject':
            self.__dict__[name] = value
            return
    method = class_type.__swig_setmethods__.get(name, None)
    if method:
        return method(self, value)
    if (not static):
        if _newclass:
            object.__setattr__(self, name, value)
        else:
            self.__dict__[name] = value
    else:
        raise AttributeError("You cannot add attributes to %s" % self)


def _swig_setattr(self, class_type, name, value):
    return _swig_setattr_nondynamic(self, class_type, name, value, 0)


def _swig_getattr_nondynamic(self, class_type, name, static=1):
    if (name == "thisown"):
        return self.this.own()
    method = class_type.__swig_getmethods__.get(name, None)
    if method:
        return method(self)
    if (not static):
        return object.__getattr__(self, name)
    else:
        raise AttributeError(name)

def _swig_getattr(self, class_type, name):
    return _swig_getattr_nondynamic(self, class_type, name, 0)


def _swig_repr(self):
    try:
        strthis = "proxy of " + self.this.__repr__()
    except:
        strthis = ""
    return "<%s.%s; %s >" % (self.__class__.__module__, self.__class__.__name__, strthis,)

try:
    _object = object
    _newclass = 1
except AttributeError:
    class _object:
        pass
    _newclass = 0


class State(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, State, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, State, name)
    __repr__ = _swig_repr
    __swig_setmethods__["next_in"] = _bzlib.State_next_in_set
    __swig_getmethods__["next_in"] = _bzlib.State_next_in_get
    if _newclass:
        next_in = _swig_property(_bzlib.State_next_in_get, _bzlib.State_next_in_set)
    __swig_setmethods__["avail_in"] = _bzlib.State_avail_in_set
    __swig_getmethods__["avail_in"] = _bzlib.State_avail_in_get
    if _newclass:
        avail_in = _swig_property(_bzlib.State_avail_in_get, _bzlib.State_avail_in_set)
    __swig_setmethods__["next_out"] = _bzlib.State_next_out_set
    __swig_getmethods__["next_out"] = _bzlib.State_next_out_get
    if _newclass:
        next_out = _swig_property(_bzlib.State_next_out_get, _bzlib.State_next_out_set)
    __swig_setmethods__["avail_out"] = _bzlib.State_avail_out_set
    __swig_getmethods__["avail_out"] = _bzlib.State_avail_out_get
    if _newclass:
        avail_out = _swig_property(_bzlib.State_avail_out_get, _bzlib.State_avail_out_set)
    __swig_setmethods__["state"] = _bzlib.State_state_set
    __swig_getmethods__["state"] = _bzlib.State_state_get
    if _newclass:
        state = _swig_property(_bzlib.State_state_get, _bzlib.State_state_set)

    def __init__(self):
        this = _bzlib.new_State()
        try:
            self.this.append(this)
        except:
            self.this = this
    __swig_destroy__ = _bzlib.delete_State
    __del__ = lambda self: None
State_swigregister = _bzlib.State_swigregister
State_swigregister(State)


def BZ2_SetInBuffer(s, data):
    return _bzlib.BZ2_SetInBuffer(s, data)
BZ2_SetInBuffer = _bzlib.BZ2_SetInBuffer

def BZ2_SetOutBuffer(s, data):
    return _bzlib.BZ2_SetOutBuffer(s, data)
BZ2_SetOutBuffer = _bzlib.BZ2_SetOutBuffer

def BZ2_Init():
    return _bzlib.BZ2_Init()
BZ2_Init = _bzlib.BZ2_Init

def BZ2_Free(s):
    return _bzlib.BZ2_Free(s)
BZ2_Free = _bzlib.BZ2_Free

def BZ2_Decompress(s):
    return _bzlib.BZ2_Decompress(s)
BZ2_Decompress = _bzlib.BZ2_Decompress

_OUTBUFSIZE = 0x1000

class BzException(Exception):
  pass

def decompress(data):
  try:
    state = BZ2_Init()
    BZ2_SetInBuffer(state, data)

    outbuf = bytearray(_OUTBUFSIZE)
    out = bytearray()

    while True:
      BZ2_SetOutBuffer(state, outbuf)
      out1 = state.next_out

      err = BZ2_Decompress(state)
      processed = int(state.next_out) - int(out1)
      if err < 0:
        break

      if processed == 0:
        break

      out += outbuf[:processed]
  finally:
    BZ2_Free(state)

  return out

# This file is compatible with both classic and new-style classes.





import os
def install(ida_dir=None):
    self_dir = os.path.dirname(os.path.abspath(__file__))
    if ida_dir is None:
        # Should be the root of the venv.
        ida_dir = os.path.join(self_dir, '../../../../../..')
    os.symlink(os.path.join(self_dir, 'loader.py'), \
               os.path.join(ida_dir, 'loaders/nsis.py'))
    os.symlink(os.path.join(self_dir, 'proc.py'), \
               os.path.join(ida_dir, 'procs/nsis_script.py'))



#!/usr/bin/env python

from setuptools import setup, Extension

bzlib = Extension('nrs.ext._bzlib', [
    'nrs/ext/bzlib.i',
    'nrs/ext/bzlib.c',
    'nrs/ext/decompress.c',
    'nrs/ext/huffman.c'
], depends=['nrs/ext/bzlib.h'])

setup(name='nrs',
      version='0.1.4',
      description='NSIS Reversing Suite',
      author='isra17',
      author_email='isra017@gmail.com',
      url='https://github.com/isra17/nrs',
      packages=['nrs','nrs.ext', 'nrs.ida', 'nrs.strings'],

      install_requires=['future'],
      setup_requires=['pytest-runner'],
      tests_require=['pytest'],
      ext_modules=[bzlib]
    )




import re
from builtins import bytes
from . import fileform, strings

from .fileform import NB_BGFONT, NB_DATA, NB_PAGES, NB_ENTRIES, NB_ENTRIES, \
                      NB_STRINGS, NB_SECTIONS, NB_CTLCOLORS, NB_LANGTABLES

def _flatten(l):
    return [i for sl in l for i in sl]

class HeaderNotFound(Exception):
    pass

class NSIS:
    @staticmethod
    def from_path(path):
        with open(path, 'rb') as fd:
            return NSIS(fd)

    def __init__(self, fd):
        """
        Create a new NSIS instance given an NSIS installer loaded in |fd|.
        """
        self._block_cache = {}
        self._pe = None

        self.fd = fd
        """ Parsed installer file. """

        self.firstheader = None
        """ Firstheader structure found at the beginning of the NSIS blob. """

        self.header = None
        """
        Header structure found at the beginning of the uncompressed NSIS blob.
        """

        self.sections = []
        """ List of sections installable by the installer. """

        self.entries = []
        """ Installer instructions. """

        self.pages = []
        """ Installer pages. """

        if not self._parse():
            raise HeaderNotFound()

        self.version_major, self.version_minor = self._detect_version()

    def get_string(self, address):
        """ Returns an NSIS expanded string given its |address|. """
        return self._parse_string(address)[0]

    def get_raw_string(self, address):
        """ Returns a raw NSIS string given its |address|. """
        string = bytearray()
        for c in self.block(NB_STRINGS)[address:]:
            if c == 0:
                break
            string.append(c)
        return string

    def get_all_strings(self):
        """ Returns all NSIS strings extracted from the strings section. """
        string_block_size = len(self.block(NB_STRINGS))
        offset = 0
        strings = []
        while offset < string_block_size:
            string, processed = self._parse_string(offset)
            if string:
                strings.append(string)
            offset += processed

        return strings

    def get_all_raw_strings(self):
        """
        Returns all raw NSIS strings extracted from the strings section.
        """
        string_block_size = len(self.block(NB_STRINGS))
        offset = 0
        strings = []
        while offset < string_block_size:
            string = self.get_raw_string(offset)
            if string:
                strings.append(string)
            offset += len(string) + 1

        return strings

    def block(self, n):
        """ Return a block data given a NB_* enum |n| value. """
        if n not in self._block_cache:
            start = self.header.blocks[n].offset
            try:
                end = next(b.offset for b
                        in self.header.blocks[n+1:] if b.offset > 0)
            except StopIteration:
                end = len(self.header.blocks)
            self._block_cache[n] = self.firstheader._raw_header[start:end]
        return self._block_cache[n]

    def size(self):
        return len(self.firstheader._raw_header)

    def close(self):
        if self._pe is not None:
            self._pe.close()

    def _detect_version(self):
        # Try to parse string and get
        nsis2_codes = 0
        nsis3_codes = 0
        for string in self.get_all_raw_strings():
            c = string[0]
            if c <= 4:
                nsis3_codes += 1
            elif c >= 252:
                nsis2_codes += 1

        if nsis2_codes > nsis3_codes:
            return '2', '?'
        else:
            return '3', '?'

    def _parse_string(self, address):
        """ Returns an NSIS expanded string given its |address|. """
        return strings.decode(self.block(NB_STRINGS), address)

    def _parse(self):
        self.firstheader = fileform._find_firstheader(self.fd)
        if self.firstheader is None:
            return False

        self.header = fileform._extract_header(self.fd, self.firstheader)

        self.pages = fileform._parse_pages(
                self.block(NB_PAGES),
                self.header.blocks[NB_PAGES].num)

        self.sections = fileform._parse_sections(
                self.block(NB_SECTIONS),
                self.header.blocks[NB_SECTIONS].num)

        self.entries = fileform._parse_entries(
                self.block(NB_ENTRIES),
                self.header.blocks[NB_ENTRIES].num)
        return True



from builtins import bytes
import struct
import zlib
from collections import namedtuple

# First header flags.
FH_FLAGS_UNINSTALL = 1
FH_FLAGS_SILENT = 2
FH_FLAGS_NO_CRC = 4
FH_FLAGS_FORCE_CRC = 8

# First header signature.
FH_SIG = 0xDEADBEEF
FH_MAGICS = b'NullsoftInst'

# Common flags.
CH_FLAGS_DETAILS_SHOWDETAILS = 1
CH_FLAGS_DETAILS_NEVERSHOW = 2
CH_FLAGS_PROGRESS_COLORED = 4
CH_FLAGS_SILENT = 8
CH_FLAGS_SILENT_LOG = 16
CH_FLAGS_AUTO_CLOSE = 32
CH_FLAGS_DIR_NO_SHOW = 64
CH_FLAGS_NO_ROOT_DIR = 128
CH_FLAGS_COMP_ONLY_ON_CUSTOM = 256
CH_FLAGS_NO_CUSTOM = 512

# Block type enumaration.
NB_PAGES = 0
NB_SECTIONS = 1
NB_ENTRIES = 2
NB_STRINGS = 3
NB_LANGTABLES = 4
NB_CTLCOLORS = 5
NB_BGFONT = 6
NB_DATA = 7

# Callback enumeration.
CB_ONINIT = 0
CB_ONINSTSUCCESS = 1
CB_ONINSTFAILED = 2
CB_ONUSERABORT = 3
#ifdef NSIS_CONFIG_ENHANCEDUI_SUPPORT
CB_ONGUIINIT = 4
CB_ONGUIEND = 5
CB_ONMOUSEOVERSECTION = 6
#endif NSIS_CONFIG_ENHANCEDUI_SUPPORT
CB_ONVERIFYINSTDIR = 7
#ifdef NSIS_CONFIG_COMPONENTPAGE
CB_ONSELCHANGE = 8
#endif NSIS_CONFIG_COMPONENTPAGE
#ifdef NSIS_SUPPORT_REBOOT
CB_ONREBOOTFAILED = 9
#endif NSIS_SUPPORT_REBOOT

# Section flags.
SF_SELECTED = 1
SF_SECGRP = 2
SF_SECGRPEND = 4
SF_BOLD = 8
SF_RO = 16
SF_EXPAND = 32
SF_PSELECTED = 64
SF_TOGGLED = 128
SF_NAMECHG = 256

# Page window proc.
#ifdef NSIS_CONFIG_LICENSEPAGE
PWP_LICENSE = 0
#endif NSIS_CONFIG_LICENSEPAGE
#ifdef NSIS_CONFIG_COMPONENTPAGE
PWP_SELCOM = 1
#endif NSIS_CONFIG_COMPONENTPAGE
PWP_DIR = 2
PWP_INSTFILES = 3
#ifdef NSIS_CONFIG_UNINSTALL_SUPPORT
PWP_UNINST = 4
#endif NSIS_CONFIG_UNINSTALL_SUPPORT
PWP_COMPLETED = 5
PWP_CUSTOM = 6

# Page flags.
PF_LICENSE_SELECTED = 1
PF_NEXT_ENABLE = 2
PF_CANCEL_ENABLE = 4
PF_BACK_SHOW = 8
PF_LICENSE_STREAM = 16
PF_LICENSE_FORCE_SELECTION = 32
PF_LICENSE_NO_FORCE_SELECTION = 64
PF_NO_NEXT_FOCUS = 128
PF_BACK_ENABLE = 256
PF_PAGE_EX = 512
PF_DIR_NO_BTN_DISABLE = 1024

# Text and background color.
CC_TEXT = 1
CC_TEXT_SYS = 2
CC_BK = 4
CC_BK_SYS = 8
CC_BKB = 16

# Delete flags.
DEL_DIR = 1
DEL_RECURSE = 2
DEL_REBOOT = 4
DEL_SIMPLE = 8

NSIS_MAX_STRLEN = 1024
NSIS_MAX_INST_TYPES = 32

MAX_ENTRY_OFFSETS = 6

BLOCKS_COUNT = 8

# First header with magic constant found in any NSIS executable.
class FirstHeader(namedtuple('FirstHeader', ['flags', 'siginfo', 'magics',
                                             'u_size', 'c_size'])):
    header_offset = 0
    data_offset = 0
    header = None

# Compressed header with the installer's sections and properties.
class Header(namedtuple('Header', [
            'flags',
            'raw_blocks',
            'install_reg_rootkey',
            'install_reg_key_ptr', 'install_reg_value_ptr',
            #ifdef NSIS_SUPPORT_BGBG
                'bg_color1s',
                'bg_color2',
                'bg_textcolor',
            #ifdef NSIS_CONFIG_VISIBLE_SUPPORT
                'lb_bg',
                'lb_fg',
            'langtable_size',
            #ifdef NSIS_CONFIG_LICENSEPAGE
                'license_bg',
            #ifdef NSIS_SUPPORT_CODECALLBACKS
                'code_onInit',
                'code_onInstSuccess',
                'code_onInstFailed',
                'code_onUserAbort',
                #ifdef NSIS_CONFIG_ENHANCEDUI_SUPPORT
                    'code_onGUIInit',
                    'code_onGUIEnd',
                    'code_onMouseOverSection',
                'code_onVerifyInstDir',
                #ifdef NSIS_CONFIG_COMPONENTPAGE
                    'code_onSelChange',
                #ifdef NSIS_SUPPORT_REBOOT
                    'code_onRebootFailed',
            #ifdef NSIS_CONFIG_COMPONENTPAGE
                'raw_install_types', # int * 32 + 1
            'install_directory_ptr',
            'install_directory_auto_append',
            #ifdef NSIS_CONFIG_UNINSTALL_SUPPORT
                'str_uninstchild',
                'str_uninstcmd',
            #ifdef NSIS_SUPPORT_MOVEONREBOOT
                'str_wininit'
        ])):
    blocks = []
    install_types = []

# Block header with location and size.
BlockHeader = namedtuple('BlockHeader', 'offset num')

Section = namedtuple('Section', [
        'name_ptr', # Initial name pointer.
        'install_types', # Bitset for the install types.
        'flags', # Flags from SF_*.
        'code', # Code location.
        'code_size', # Size of the code.
        'size_kb',
        'name' # Empty for invisible sections.
    ])

class Entry(namedtuple('Entry', [
            'which', # EW_* enum.
            'raw_offsets', # Meaning depends on |which|.
        ])):
    offsets = []


class Page(namedtuple('Page', [
            'dlg_id', # Dialog resource ID.
            'wndproc_id',
            #ifdef NSIS_SUPPORT_CODECALLBACKS
                'prefunc', # Called before the page is created.
                'showfunc', # Called right before the page is shown.
                'leavefunc', # Called when the user leaves the page.
            'flags',
            'caption',
            'back',
            'next',
            'clicknext',
            'cancel',
            'raw_params'
        ])):
    params = []

CtlColors32 = namedtuple('CtlColors32', [
        'text',
        'bkc',
        'lbStyle',
        'bkb',
        'bkmode',
        'flags'
    ])


_firstheader_pack = struct.Struct("<II12sII")
_header_pack = struct.Struct("<I64s20I{}s5I".format(4*(NSIS_MAX_INST_TYPES+1)))
_blockheader_pack = struct.Struct("<II")
_section_pack = struct.Struct("<6I{}s".format(NSIS_MAX_STRLEN))
_entry_pack = struct.Struct("<I{}s".format(MAX_ENTRY_OFFSETS*4))
_page_pack = struct.Struct("<11I20s")
_ctlcolors32_pack = struct.Struct("<6I")

def _find_firstheader(nsis_file):
    firstheader_offset = 0
    pos = 0
    while True:
        chunk = nsis_file.read(32768 if firstheader_offset else 512)
        if len(chunk) < _firstheader_pack.size:
            return None

        if firstheader_offset == 0:
            firstheader = FirstHeader._make(
                    _firstheader_pack.unpack_from(chunk))
            firstheader.header_offset = pos
            firstheader.data_offset = pos + _firstheader_pack.size

            if firstheader.siginfo == FH_SIG and \
                    firstheader.magics == FH_MAGICS:
                # NSIS header found.
                return firstheader

        pos += len(chunk)

def _is_lzma(data):
    def _is_lzma_header(data):
        return data[0:3] == [0x5d, 0, 0] and data[5] == 0 and (data[6] & 0x80 == 0)
    return (_is_lzma_header(data) or (data[0] <= 1 and _is_lzma_header(data[1:])))

def _is_bzip2(data):
    return data[0] == 0x31 and data[1] < 0xe

def _zlib(f, size):
    data = f.read(size)
    return zlib.decompress(data, -zlib.MAX_WBITS)

def _bzip2(f, size):
    from nrs.ext import bzlib
    data = f.read()
    inflated_data = bytes(bzlib.decompress(data))
    size, = struct.unpack_from('<I', inflated_data)
    return inflated_data[4:size+4]

def inflate_header(nsis_file, data_offset):
    nsis_file.seek(data_offset)
    chunk = bytes(nsis_file.read(0xc))
    data_size = struct.unpack_from('<I', chunk)[0]
    solid = True
    decoder = None

    if _is_lzma(chunk):
        raise NotImplemented('Solid LZMA')
    elif chunk[3] == 0x80:
        solid = False
        if _is_lzma(chunk[4:]):
            raise NotImplemented('Not-Solid LZMA')
        elif _is_bzip2(chunk[4:]):
            raise NotImplemented('Not-Solid Bzip2')
        else:
            decoder = _zlib
    elif _is_bzip2(chunk):
        decoder = _bzip2
    else:
        raise NotImplemented('Solid Deflate')

    if solid:
        deflated_data = nsis_file.seek(data_offset)
    else:
        nsis_file.seek(data_offset+4)
        data_size &= 0x7fffffff

    inflated_data = decoder(nsis_file, data_size)

    return inflated_data, data_size

def _extract_header(nsis_file, firstheader):
    inflated_data, data_size = inflate_header(nsis_file, firstheader.data_offset)
    header = Header._make(_header_pack.unpack_from(inflated_data))
    firstheader.header = header
    firstheader._raw_header = bytes(inflated_data)
    firstheader._raw_header_c_size = data_size

    # Parse the block headers.
    block_headers = []
    for i in range(BLOCKS_COUNT):
        header_offset = i * _blockheader_pack.size
        block_header = BlockHeader._make(_blockheader_pack.unpack_from(
            header.raw_blocks[header_offset:]))
        block_headers.append(block_header)
    header.blocks = block_headers

    # Parse the install types.
    header.install_types = [
            struct.unpack_from('<I', header.raw_install_types[i:])[0]
                for i in range(0, len(header.raw_install_types), 4)]

    return header

def _extract_block(nsis_file, firstheader, block_id):
    header = firstheader.header
    if block_id == NB_DATA:
        nsis_file.seek(firstheader.data_offset + firstheader._raw_header_c_size)
        return nsis_file.read()

    return firstheader._raw_header[header.blocks[block_id].offset:]

def _parse_sections(block, n):
    bsize = _section_pack.size
    return [Section._make(_section_pack.unpack_from(block[i * bsize:]))
                for i in range(n)]

def _parse_entries(block, n):
    bsize = _entry_pack.size
    entries = []
    for i in range(n):
        entry = Entry._make(_entry_pack.unpack_from(block[i * bsize:]))
        # Parse the install types.
        entry.offsets = [
            struct.unpack_from('<I', entry.raw_offsets[j:])[0]
                for j in range(0, len(entry.raw_offsets), 4)]
        entries.append(entry)

    return entries

def _parse_pages(block, n):
    bsize = _page_pack.size
    pages = []
    for i in range(n):
        page = Page._make(_page_pack.unpack_from(block[i * bsize:]))
        # Parse the install types.
        page.params = [
            struct.unpack_from('<I', page.raw_params[j:])[0]
                for j in range(0, len(page.raw_params), 4)]
        pages.append(page)

    return pages

